// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2017, Mellanox Technologies. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/acpi.h>
#include <linux/device.h>
#include <linux/resource.h>
#include <linux/bitfield.h>
#include <linux/bitops.h>
#include <linux/clk.h>
#include <linux/io.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/math64.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/pinctrl/consumer.h>
#include <linux/pm.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/platform_device.h>
#include <linux/string.h>
#include <linux/of_address.h>
#include <linux/reboot.h>
#include <linux/gpio/driver.h>
#include <linux/version.h>

#include "gpio-mlxbf_regs.h"

/* Number of pins on BlueField */
#define BF_GPIO_NR 54

/* GPIO pin 7 is reserved for power good indication on BlueField. */
#define BF_GPIO_POWER_GOOD_PIN	7

/* GPIO pins enabled for interrupt */
#define BF_GPIO_INT_ASSERT_ENA_BITS	0
#define BF_GPIO_INT_DEASSERT_ENA_BITS	BIT(BF_GPIO_POWER_GOOD_PIN)

/* Pad Electrical Controls. */
#define GPIO_PAD_CONTROL__FIRST_WORD 0x0700
#define GPIO_PAD_CONTROL_1__FIRST_WORD 0x0708
#define GPIO_PAD_CONTROL_2__FIRST_WORD 0x0710
#define GPIO_PAD_CONTROL_3__FIRST_WORD 0x0718

#ifdef CONFIG_PM
struct bluefield_context_save_regs {
	uint64_t gpio_scratchpad;
	uint64_t gpio_pad_control[BF_GPIO_NR];
	uint64_t gpio_pin_dir_i;
	uint64_t gpio_pin_dir_o;
};
#endif


/** Device state structure. */
struct gpio_state {
	struct gpio_chip gc;

	struct list_head node;
	struct platform_device *pdev;

	/* Must hold this lock to modify shared data. */
	spinlock_t lock;

	/* Memory Address */
	void __iomem *dc_base;

	/* Interrupt number */
	u16 hwirq;

#ifdef CONFIG_PM
	struct bluefield_context_save_regs csave_regs;
#endif
};

void acpi_gpiochip_request_interrupts(struct gpio_chip *chip);

/* GPIO driver set input pins */
static int gpio_bf_set_input(struct gpio_chip *chip, unsigned int offset)
{
	struct gpio_state *gs = gpiochip_get_data(chip);
	uint64_t in;
	uint64_t out;

	if (offset > BF_GPIO_NR - 1) {
		printk(KERN_ERR "gpio input pins: Invalid offset %d\n",
			offset);
		return -EINVAL;
	}

	out = readq(gs->dc_base + GPIO_PIN_DIR_O);
	in = readq(gs->dc_base + GPIO_PIN_DIR_I);

	writeq(out & ~(1ULL << offset), gs->dc_base + GPIO_PIN_DIR_O);
	writeq(in | (1ULL << offset), gs->dc_base + GPIO_PIN_DIR_I);

	return 0;
}

/* GPIO driver set output pins */
static int gpio_bf_set_output(struct gpio_chip *chip, unsigned int offset)
{
	struct gpio_state *gs = gpiochip_get_data(chip);
	uint64_t in;
	uint64_t out;

	if (offset > BF_GPIO_NR - 1) {
		printk(KERN_ERR "gpio output pins: Invalid offset %d\n",
			offset);
		return -EINVAL;
	}

	out = readq(gs->dc_base + GPIO_PIN_DIR_O);
	in = readq(gs->dc_base + GPIO_PIN_DIR_I);

	writeq(out | (1ULL << offset), gs->dc_base + GPIO_PIN_DIR_O);
	writeq(in & ~(1ULL << offset), gs->dc_base + GPIO_PIN_DIR_I);

	return 0;
}

/* GPIO driver output direction */
static int gpio_bf_set_output_lock(struct gpio_chip *chip,
		unsigned int offset, int value)
{
	struct gpio_state *gs = gpiochip_get_data(chip);

	spin_lock(&gs->lock);
	gpio_bf_set_output(chip, offset);
	spin_unlock(&gs->lock);

	return 0;
}


/* GPIO driver input direction */
static int gpio_bf_set_input_lock(struct gpio_chip *chip,
		unsigned int offset)
{
	struct gpio_state *gs = gpiochip_get_data(chip);

	spin_lock(&gs->lock);
	gpio_bf_set_input(chip, offset);
	spin_unlock(&gs->lock);

	return 0;
}

/* GPIO driver read routine. */
static int gpio_bf_get(struct gpio_chip *chip, unsigned int offset)
{
	uint64_t value;
	struct gpio_state *gs = gpiochip_get_data(chip);

	value = readq(gs->dc_base + GPIO_PIN_STATE);

	return (value >> offset) & 1;
}

/* GPIO driver write routine. */
static void gpio_bf_set(struct gpio_chip *chip, unsigned int offset,
	int value)
{
	uint64_t data;
	struct gpio_state *gs = gpiochip_get_data(chip);

	spin_lock(&gs->lock);
	data = readq(gs->dc_base + GPIO_PIN_STATE);

	if (value)
		data |= (1ULL << offset);
	else
		data &= ~(1ULL << offset);
	writeq(data, gs->dc_base + GPIO_PIN_STATE);
	spin_unlock(&gs->lock);
}

static int gpio_bf_to_irq(struct gpio_chip *chip, u32 gpio)
{
	struct gpio_state *gs = gpiochip_get_data(chip);

	return gs->hwirq;
}

static void gpio_bf_cfg_irq(struct gpio_state *gs)
{
	u64 intr_cfg;

	intr_cfg = readq(gs->dc_base + GPIO_INT_SETUP);
	intr_cfg |= FIELD_PREP(GPIO_INT_SETUP__GBL_ENA_MASK, 1);
	writeq(intr_cfg, gs->dc_base + GPIO_INT_SETUP);

	intr_cfg = readq(gs->dc_base + GPIO_INT_ENA_ASSERT);
	intr_cfg |= FIELD_PREP(GPIO_INT_ENA_ASSERT__INT_ENA_ASSERT_MASK,
			       BF_GPIO_INT_ASSERT_ENA_BITS);
	writeq(intr_cfg, gs->dc_base + GPIO_INT_ENA_ASSERT);

	intr_cfg = readq(gs->dc_base + GPIO_INT_ENA_DEASSERT);
	intr_cfg |= FIELD_PREP(GPIO_INT_ENA_DEASSERT__INT_ENA_DEASSERT_MASK,
			       BF_GPIO_INT_DEASSERT_ENA_BITS);
	writeq(intr_cfg, gs->dc_base + GPIO_INT_ENA_DEASSERT);
}

/* GPIO driver initialization routine. */
static int
gpiodrv_probe(struct platform_device *pdev)
{
	struct gpio_state *gs;
	struct device *dev = &pdev->dev;
	struct gpio_chip *gc;
	struct resource *dc_res;
	int ret, irq;

	dc_res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (dc_res == NULL)
		return -EINVAL;

	irq = platform_get_irq(pdev, 0);
	if (irq < 0)
		return irq;

	if (request_mem_region(dc_res->start,
			       resource_size(dc_res), "gpiodrv") == NULL)
		return -EBUSY;

	ret = -ENOMEM;

	gs = kzalloc(sizeof(struct gpio_state), GFP_KERNEL);
	if (gs == NULL)
		goto err_release;

	gs->dc_base = ioremap(dc_res->start, resource_size(dc_res));
	if (!gs->dc_base)
		goto err_free;

	gc = &gs->gc;
	gc->direction_input = gpio_bf_set_input_lock;
	gc->get = gpio_bf_get;
	gc->direction_output = gpio_bf_set_output_lock;
	gc->set = gpio_bf_set;
	gc->label = dev_name(dev);
	gc->parent = &pdev->dev;
	gc->owner = THIS_MODULE;
	gc->base = 0;
	gc->ngpio = BF_GPIO_NR;
	gs->hwirq = irq;
	gc->to_irq = gpio_bf_to_irq;

	gpio_bf_cfg_irq(gs);

	ret = gpiochip_add_data(&gs->gc, gs);
	if (ret) {
		dev_err(&pdev->dev, "Failed adding memory mapped gpiochip\n");
		goto err;
	}

	spin_lock_init(&gs->lock);
	platform_set_drvdata(pdev, gs);
	dev_info(&pdev->dev, "registered Mellanox BlueField gpio");

	acpi_gpiochip_request_interrupts(gc);

	return 0;

err:
	iounmap(gs->dc_base);
err_free:
	kfree(gs);
err_release:
	release_mem_region(dc_res->start, resource_size(dc_res));
	dev_err(&pdev->dev, "Probe, Failed\n");
	return ret;
}

#ifdef CONFIG_PM
static int gpiodrv_suspend(struct platform_device *pdev,
		pm_message_t state)
{
	struct gpio_state *gs = platform_get_drvdata(pdev);

	gs->csave_regs.gpio_scratchpad = readq(gs->dc_base +
	    GPIO_SCRATCHPAD);

	gs->csave_regs.gpio_pad_control[0] = readq(gs->dc_base +
	    GPIO_PAD_CONTROL__FIRST_WORD);
	gs->csave_regs.gpio_pad_control[1] = readq(gs->dc_base +
	    GPIO_PAD_CONTROL_1__FIRST_WORD);
	gs->csave_regs.gpio_pad_control[2] = readq(gs->dc_base +
	    GPIO_PAD_CONTROL_2__FIRST_WORD);
	gs->csave_regs.gpio_pad_control[3] = readq(gs->dc_base +
	    GPIO_PAD_CONTROL_3__FIRST_WORD);
	gs->csave_regs.gpio_pin_dir_i = readq(gs->dc_base +
	    GPIO_PIN_DIR_I);
	gs->csave_regs.gpio_pin_dir_o = readq(gs->dc_base +
	    GPIO_PIN_DIR_O);

	return 0;
}

static int gpiodrv_resume(struct platform_device *pdev)
{
	struct gpio_state *gs = platform_get_drvdata(pdev);

	writeq(gs->csave_regs.gpio_scratchpad, gs->dc_base + GPIO_SCRATCHPAD);
	writeq(gs->csave_regs.gpio_pad_control[0], gs->dc_base +
	    GPIO_PAD_CONTROL__FIRST_WORD);
	writeq(gs->csave_regs.gpio_pad_control[1], gs->dc_base +
	    GPIO_PAD_CONTROL_1__FIRST_WORD);
	writeq(gs->csave_regs.gpio_pad_control[2], gs->dc_base +
	    GPIO_PAD_CONTROL_2__FIRST_WORD);
	writeq(gs->csave_regs.gpio_pad_control[3], gs->dc_base +
	    GPIO_PAD_CONTROL_3__FIRST_WORD);
	writeq(gs->csave_regs.gpio_pin_dir_i, gs->dc_base + GPIO_PIN_DIR_I);
	writeq(gs->csave_regs.gpio_pin_dir_o, gs->dc_base + GPIO_PIN_DIR_O);

	return 0;
}
#endif

static int gpiodrv_remove(struct platform_device *pdev)
{
	struct gpio_state *gs = platform_get_drvdata(pdev);
	struct resource *dc_res =
		platform_get_resource(pdev, IORESOURCE_MEM, 0);

	gpiochip_remove(&gs->gc);

	iounmap(gs->dc_base);

	kfree(gs);

	release_mem_region(dc_res->start, resource_size(dc_res));

	return 0;
}

static const struct of_device_id gpiodrv_gpio_match[] = {
	{ .compatible = "mellanox,mlxbf-gpio" },
	{},
};

MODULE_DEVICE_TABLE(of, gpiodrv_gpio_match);

static const struct acpi_device_id gpiodrv_acpi_match[] = {
	{ "MLNXBF02", 0 },
	{},
};
MODULE_DEVICE_TABLE(acpi, gpiodrv_acpi_match);

static struct platform_driver gpiodrv_gpio_driver = {
	.driver = {
		.name = "gpiodrv",
		.of_match_table = gpiodrv_gpio_match,
		.acpi_match_table = ACPI_PTR(gpiodrv_acpi_match),
	},
	.probe    = gpiodrv_probe,
	.remove   = gpiodrv_remove,
#ifdef CONFIG_PM
	.suspend  = gpiodrv_suspend,
	.resume   = gpiodrv_resume,
#endif
};

module_platform_driver(gpiodrv_gpio_driver);

MODULE_DESCRIPTION("Mellanox BlueField GPIO Driver");
MODULE_AUTHOR("Mellanox Technologies");
MODULE_LICENSE("GPL");
